/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Simple Timer + Clocks.
 *
 * The Initial Developer of the Original Code is
 *   George Bradt.
 *
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Credits:
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

var SimpleClocksCalendar = {
    // For reference, not used.
    calEvent: {recurring: false,
               startDateDisplay: "",
               startDate: "",
               startTimeMilli: 0,
               endDateDisplay: "",
               endDate: "",
               endTimeMilli: 0,
               timeDisplay: "",
               time24hr: "",
               repeatRule: {freq: -1, byDay: ""},
               untilDateDisplay: "",
               untilDate: "",
               untilTimeMilli: 0,
               alertDays: 0,
               alertMilli: 0,
               description: ""
              },
    calEventSep: "\x1D", // used to separate nots. in pref.

    // Called when loading Calendar Events dialog.

    onLoad: function() {
        var calEventTable = this.getCalEventTable("calEventTable");

        if ( calEventTable ) {
            var treeChildren = document.getElementById("simclk-treeItems");

            for ( var i in calEventTable ) {
                this.addTreeRow(treeChildren, calEventTable[i]);
            }
        }

        this.setFreqLabels();
    },

    // The change event is fired whenever the start date is changed.
    // Change end, until dates as well.

    onChangeStartDate: function() {
        var startDate = document.getElementById("simclk-dpickEventStartDate").value;

        document.getElementById("simclk-dpickEventEndDate").value = startDate;
        document.getElementById("simclk-dpickEventUntilDate").value = startDate;

        this.setFreqLabels();
    },

    // Called when user clicks Today button.
    // Set datepickers to current date.

    onToday: function() {
        var currDate = new Date();

        document.getElementById("simclk-dpickEventStartDate").dateValue = currDate;
        document.getElementById("simclk-dpickEventEndDate").dateValue = currDate;
        document.getElementById("simclk-dpickEventUntilDate").dateValue = currDate;
    },

    // Called when user clicks All day checkbox.
    // Enable/disable timepicker.

    onClickAllDay: function() {
        var timePicker = document.getElementById("simclk-tpickEventTime");

        timePicker.value = "00:00:00";
        timePicker.disabled = document.getElementById("simclk-cboxEventAllDay").checked;
    },

    // Called when user clicks Recurring checkbox.
    // Enable/disable stuff.

    onClickRecurring: function() {
        var recurring = document.getElementById("simclk-cboxEventRecurring").checked;

        document.getElementById("simclk-mlistFreq").disabled = !recurring;
        document.getElementById("simclk-dpickEventUntilDate").disabled = !recurring;
        document.getElementById("simclk-cboxForever").disabled = !recurring;

        if ( !recurring ) {
            document.getElementById("simclk-cboxForever").checked = false;
        }
    },

    // Set the menuitem labels for the date format menulist.
    // Called on dialog load and when start date changes.

    setFreqLabels: function() {
        var strbundle = document.getElementById("simclk-strings");
        var startDate = document.getElementById("simclk-dpickEventStartDate").dateValue;
        var date = startDate.getDate();
        var month = startDate.getMonth();
        var year = startDate.getFullYear();
        var nth = this.dateToNthDay(date, month, year);
        var byDay = nth + this.getByDay(startDate.getDay());
        var nthStr = this.getNthDow(nth);
        var dowStr = startDate.toLocaleFormat("%A");
        var monthStr = startDate.toLocaleFormat("%B");

        document.getElementById("simclk-mitemFreq0").label =
                strbundle.getString("display.freq.yearly");
        document.getElementById("simclk-mitemFreq0").value = "";

        document.getElementById("simclk-mitemFreq1").label =
                strbundle.getString("display.freq.yearly.by.day") + " " +
                strbundle.getString("display.freq.example").
                        replace("%nth", nthStr).replace("%dow", dowStr).replace("%mon", monthStr);
        document.getElementById("simclk-mitemFreq1").value = byDay;

        document.getElementById("simclk-mitemFreq2").label =
                strbundle.getString("display.freq.monthly");
        document.getElementById("simclk-mitemFreq2").value = "";

        document.getElementById("simclk-mitemFreq3").label =
                strbundle.getString("display.freq.monthly.by.day") + " " +
                strbundle.getString("display.freq.example").
                        replace("%nth", nthStr).replace("%dow", dowStr).
                        replace("%mon", strbundle.getString("display.month"));
        document.getElementById("simclk-mitemFreq3").value = byDay;

        document.getElementById("simclk-mitemFreq4").label =
                strbundle.getString("display.freq.weekly");
        document.getElementById("simclk-mitemFreq4").value = "";

        document.getElementById("simclk-mitemFreq5").label =
                strbundle.getString("display.dow.daily");
        document.getElementById("simclk-mitemFreq5").value = "";
    },

    // eg. August 19, 2009 ==> third Wednesday in August.

    dateToNthDay: function (date, month, year) {
        var nth = 1;
        var testDate = new Date(year, month, date);

        testDate.setDate(date + 7);

        if ( testDate.getMonth() !== month ) {
            // Last.
            nth = -1;
        }
        else {
            testDate.setDate(date - 7);

            while ( testDate.getMonth() === month ) {
                testDate.setDate(testDate.getDate() - 7);
                nth++;
            }
        }

        return nth;
    },

    // Convert numeric value to nth day of week.

    getNthDow: function(nth) {
        var strbundle = document.getElementById("simclk-strings");
        var nthDays = [strbundle.getString("alert.error.invalid.dow"),
                       strbundle.getString("display.dow.first"),
                       strbundle.getString("display.dow.second"),
                       strbundle.getString("display.dow.third"),
                       strbundle.getString("display.dow.fourth")
                      ];

        if ( nthDays[nth] != null ) {
            // When undefined (ie. nthDays[-1]) is compared to null, the result is true.
            return nthDays[nth];
        }
        else {
            if ( nth === -1 ) {
                return strbundle.getString("display.dow.last");

            }
            else {
                alert(strbundle.getString("alert.error.invalid.dow"));
                return "Err";
            }
        }
    },

    // Convert numeric weekday to value used in ical BYDAY code.

    getByDay: function(dow) {
        var strbundle = document.getElementById("simclk-strings");
        var byDays = ["SU", "MO", "TU", "WE", "TH", "FR", "SA"];

        if ( byDays[dow] != null ) {
            // When undefined (ie. byDays[-1]) is compared to null, the result is true.
            return byDays[dow];
        }
        else {
            alert(strbundle.getString("alert.error.invalid.dow"));
            return "Err";
        }
    },


    // eg. third Wednesday in August 2009 ==> August 19, 2009.
    // nth = -1 return last occurrence.

    nthDayToDate: function (eventDate, nth, targetDow) {
        var strbundle = document.getElementById("simclk-strings");
        var month = eventDate.getMonth();

        if ( nth === -1 ) {
            // Set to last week of month.
            eventDate.setDate(1);
            eventDate.setMonth(month + 1);
            eventDate.setDate(-6);
            nth = 1;
        }
        else {
            // Set to first day of month.
            eventDate.setDate(1);
        }

        var offset = targetDow - eventDate.getDay();
        eventDate.setDate(eventDate.getDate() + (offset + (nth - (offset >= 0 ? 1 : 0)) * 7));

        if (eventDate.getMonth() !== month) {
            alert(strbundle.getString("alert.error.invalid.nth.calc") + " " + nth + " " + targetDow);
        }
    },

    // Called when user clicks Forever checkbox.
    // Enable/disable Repeat until datepicker.

    onClickForever: function() {
        document.getElementById("simclk-dpickEventUntilDate").disabled =
                document.getElementById("simclk-cboxForever").checked;
    },

    // Called when user clicks on Events tree.

    onClickEventTree: function(event) {
        var strbundle = document.getElementById("simclk-strings");
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);
        var tree = document.getElementById("simclk-treeEvents");

        // See if tree disabled (ie. edit already in progress).
        if ( tree.disabled ) {
            return;
        }

        // Get the row, col and child element at the point of mouse click
        var tbo = tree.treeBoxObject;
        var row = { }, col = { }, child = { };
        tbo.getCellAt(event.clientX, event.clientY, row, col, child);

        if ( col.value ) {
            if ( col.value.id !== "simclk-colRecur" &&
                event.detail === 2 ) {
                // No monkey business.
                tree.disabled = true;

                // User double-clicked on an event to initiate an edit.
                // First disable the delete and reset buttons.
                document.getElementById("simclk-btnDel").disabled = true;
                document.getElementById("simclk-btnDelAll").disabled = true;
                document.getElementById("simclk-btnReset").disabled = true;

                // Change label and tooltips on Add and Add/OK buttons.
                document.getElementById("simclk-btnAddOk").label =
                        strbundle.getString("button.label.replace/OK");
                document.getElementById("simclk-btnAddOk").tooltipText =
                        strbundle.getString("button.tooltip.replace/OK");
                document.getElementById("simclk-btnAdd").label =
                        strbundle.getString("button.label.replace");
                document.getElementById("simclk-btnAdd").tooltipText =
                        strbundle.getString("button.tooltip.replace");

                // Move the event data back to the upper half.
                document.getElementById("simclk-tboxEventDesc").value =
                        tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colDescription"));

                // Note that this value may not be the same as the original value,
                // it gets updated for recurring events.
                document.getElementById("simclk-dpickEventStartDate").value =
                        tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colStartDate"));

                document.getElementById("simclk-dpickEventEndDate").value =
                        tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colEndDate"));

                var timePicker = document.getElementById("simclk-tpickEventTime");
                timePicker.value = tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colTime"));

                var timeDisplay = tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colTime"));

                if ( timeDisplay === strbundle.getString("display.all.day") ) {
                    document.getElementById("simclk-cboxEventAllDay").checked = true;
                    timePicker.disabled = true;
                }
                else {
                    document.getElementById("simclk-cboxEventAllDay").checked = false;
                    timePicker.disabled = false;
                }

                document.getElementById("simclk-tboxAlertDays").value =
                        tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colAlertDays"));

                var mlistFreq = document.getElementById("simclk-mlistFreq");
                var recurring = tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colRecur"));

                var untilDatePicker = document.getElementById("simclk-dpickEventUntilDate");
                var until = json.decode(tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colUntilDate")));
                untilDatePicker.value = until.untilDate;
                var untilDateDisplay = tree.view.getCellText(row.value,
                        tree.columns.getNamedColumn("simclk-colUntilDate"));

                var cboxForever = document.getElementById("simclk-cboxForever");

                if ( recurring === "true" ) {
                    document.getElementById("simclk-cboxEventRecurring").checked = true;
                    var repeatRule = json.decode(tree.view.getCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colFreq")));
                    mlistFreq.selectedIndex = repeatRule.freq;
                    mlistFreq.selectedItem.value = repeatRule.byDay;
                    mlistFreq.disabled = false;
                    untilDatePicker.disabled = false;
                    cboxForever.disabled = false;

                    if ( untilDateDisplay === strbundle.getString("display.forever") ) {
                        cboxForever.checked = true;
                    }
                    else {
                        cboxForever.checked = false;
                    }
                }
                else {
                    document.getElementById("simclk-cboxEventRecurring").checked = false;
                    mlistFreq.selectedIndex = 0;
                    mlistFreq.selectedItem.value = "";
                    mlistFreq.disabled = true;
                    untilDatePicker.disabled = true;
                    cboxForever.disabled = true;
                    cboxForever.checked = false;
                }

                // Stash the row in a custom attribute.
                document.getElementById("simclk-tboxEventDesc").
                        setAttribute("updaterow", row.value);

                return;
            }

            if ( col.value.id === "simclk-colRecur" ) {
                // Toggling Recurring also changes Freq and Repeat until.
                // Unless an actual until date is provided, until date = end date.
                var untilDate = tree.view.getCellValue(row.value,
                        tree.columns.getNamedColumn("simclk-colEndDate"));
                until = json.encode({untilDate: untilDate, untilTimeMilli: 0});

                // Post-click value.
                if ( tree.view.getCellValue(row.value, col.value) === "true" ) {
                    repeatRule = json.encode({freq: 0, byDay: ""});
                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colFreq"),
                            strbundle.getString("display.freq.yearly"));
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colFreq"), repeatRule);

                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colUntilDate"),
                            strbundle.getString("display.forever"));
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colUntilDate"),
                            until);
                }
                else {
                    repeatRule = json.encode({freq: -1, byDay: ""});
                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colFreq"),
                            strbundle.getString("msg.not.applicable"));
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colFreq"), repeatRule);

                    tree.view.setCellText(row.value,
                            tree.columns.getNamedColumn("simclk-colUntilDate"),
                            strbundle.getString("msg.not.applicable"));
                    tree.view.setCellValue(row.value,
                            tree.columns.getNamedColumn("simclk-colUntilDate"),
                            until);
                }

                return;
            }
        }
    },

    // Called when Add/OK button is clicked.

    onAddOk: function() {
        if ( this.onAdd() ) {
            this.onOK();
            window.close();
        }
    },

    // Called when Add button is clicked.

    onAdd: function() {
        var strbundle = document.getElementById("simclk-strings");
        var descr = document.getElementById("simclk-tboxEventDesc").value;

        if ( !descr ) {
            alert(strbundle.getString("alert.error.no.description"));
            document.getElementById("simclk-tboxEventDesc").focus();
            return false;
        }

        // Examine start date datepicker value.
        var startDatePicker = document.getElementById("simclk-dpickEventStartDate");
        var startDateStr = startDatePicker.value;
        var currDate = new Date();
        var currDateStr = currDate.getFullYear() + "-" +
                this.pad(currDate.getMonth() + 1) + "-" +
                this.pad(currDate.getDate());

        if ( currDateStr > startDateStr ) {
            // Don't allow past dates.
            alert(strbundle.getString("alert.error.bad.start.date"));
            return false;
        }

        // The dateValue property is a date object, but it uses the current time.
        var startDate = startDatePicker.dateValue;
        startDate.setHours(0,0,0,0);
        var startDateDisplay = startDate.toLocaleFormat("%b %d %Y");
        var startTimeMilli = startDate.getTime();

        // Examine end date datepicker value.
        var endDatePicker = document.getElementById("simclk-dpickEventEndDate");
        var endDateStr = endDatePicker.value;

        if ( endDateStr < startDateStr ) {
            endDatePicker.value = startDateStr;
            endDateStr = startDateStr;
            alert(strbundle.getString("alert.error.bad.end.date"));
        }

        var endDate = endDatePicker.dateValue;
        endDate.setHours(0,0,0,0);
        var endDateDisplay = endDate.toLocaleFormat("%b %d %Y");
        var endTimeMilli = endDate.getTime();

        var recurring = document.getElementById("simclk-cboxEventRecurring").checked;

        if ( recurring ) {
            // Default freq is yearly.
            var mlistFreq = document.getElementById("simclk-mlistFreq");
            var freq = mlistFreq.selectedIndex;
            var byDay = mlistFreq.value;

            // Examine repeat until datepicker value.
            if ( document.getElementById("simclk-cboxForever").checked ) {
                var untilDateStr = endDateStr;
                var untilDateDisplay = strbundle.getString("display.forever");
                var untilTimeMilli = 0;
            }
            else {
                var untilDatePicker = document.getElementById("simclk-dpickEventUntilDate");
                untilDateStr = untilDatePicker.value;

                if ( untilDateStr <= endDateStr ) {
                    if ( untilDateStr < endDateStr ) {
                        untilDatePicker.value = endDateStr;
                        untilDateStr = endDateStr;
                    }

                    alert(strbundle.getString("alert.error.bad.until.date"));
                }

                var untilDate = untilDatePicker.dateValue;
                untilDate.setHours(0,0,0,0);
                untilDateDisplay = untilDate.toLocaleFormat("%b %d %Y");
                untilTimeMilli = untilDate.getTime();
            }
        }
        else {
            freq = -1;
            byDay = "";
            untilDateStr = endDateStr;
            untilDateDisplay = strbundle.getString("msg.not.applicable");
            untilTimeMilli = 0;
        }

        // Examine timepicker value. Remove seconds from timeDisplay.
        // Note that time is for display purposes only, alerts are based on date.
        // time24hr is used for double-click editing.
        var time24hr = document.getElementById("simclk-tpickEventTime").value;
        var timeDisplay = ( document.getElementById("simclk-cboxEventAllDay").checked === true ) ?
                strbundle.getString("display.all.day") :
                new Date("December 23, 2007 " + time24hr).toLocaleFormat("%X").replace(/:\d\d(?!.*:\d\d)/, "");

        var alertDays = document.getElementById("simclk-tboxAlertDays").value;
        var alertMilli = 0;

        if ( alertDays > 0 ) {
            // We can change startDate as its not needed anymore.
            startDate.setDate(startDate.getDate() - alertDays);
            alertMilli = startDate.getTime();
        }

        var tree = document.getElementById("simclk-treeEvents");
        var updaterow = document.getElementById('simclk-tboxEventDesc').
                getAttribute("updaterow");

        if ( updaterow ) {
            tree.disabled = false;

            // Updating a notification, enable all buttons.
            document.getElementById("simclk-btnDel").disabled = false;
            document.getElementById("simclk-btnDelAll").disabled = false;
            document.getElementById("simclk-btnReset").disabled = false;

            // Reset label and tooltip on Add and Add/OK buttons.
            document.getElementById("simclk-btnAddOk").label =
                    strbundle.getString("button.label.add/OK");
            document.getElementById("simclk-btnAddOk").tooltipText =
                        strbundle.getString("button.tooltip.add/OK");
            document.getElementById("simclk-btnAdd").label =
                    strbundle.getString("button.label.add");
            document.getElementById("simclk-btnAdd").tooltipText =
                        strbundle.getString("button.tooltip.add");
        }

        var data = {recurring: recurring, startDateDisplay: startDateDisplay, startDate: startDateStr,
                startTimeMilli: startTimeMilli, endDateDisplay: endDateDisplay, endDate: endDateStr,
                endTimeMilli: endTimeMilli, timeDisplay: timeDisplay, time24hr: time24hr,
                repeatRule: {freq: freq, byDay: byDay}, untilDateDisplay: untilDateDisplay,
                untilDate: untilDateStr, untilTimeMilli: untilTimeMilli, alertDays: alertDays,
                alertMilli: alertMilli, description: descr};

        var treeChildren = document.getElementById("simclk-treeItems");
        var treeItems = treeChildren.hasChildNodes() ?
                treeChildren.childNodes : null;

        this.addTreeRow(treeChildren, data);

        // Select the added/updated row.
        var rowIndex;

        if ( updaterow ) {
            rowIndex = parseInt(updaterow, 10);
        }
        else if ( treeItems ) {
            rowIndex = treeItems.length - 1;
        }

        tree.view.selection.select(rowIndex);

        var boxobject = tree.boxObject;
        boxobject.QueryInterface(Components.interfaces.nsITreeBoxObject);

        // Ensure the added/updated row is visible, tree displays 5 rows.
        if ( treeItems && treeItems.length > 5 ) {
            boxobject.ensureRowIsVisible(rowIndex);
        }

        // Clear the textbox.
        document.getElementById("simclk-tboxEventDesc").value = "";
        document.getElementById("simclk-tboxEventDesc").focus();

        return true;
    },

    // Called when Import button is clicked.

    /*onImport: function() {
        // Display the file picker dialog.
        var strbundle = document.getElementById("simclk-strings");
        var filePicker = Components.classes["@mozilla.org/filepicker;1"].
                createInstance(Components.interfaces.nsIFilePicker);

        filePicker.init(window,
                        strbundle.getString("title.ics.filepicker"),
                        filePicker.modeOpen);
        filePicker.appendFilter("*.ics", "*.ics");

        if ( filePicker.show() === filePicker.returnOK ) {
            var file = SimpleClocksFileIO.open(filePicker.file.path);

			if ( file ) {
				var icsStream = SimpleClocksFileIO.read(file);

                if ( icsStream ) {
                    SimpleClocksIcalParser.parseIcal(icsStream);
                }
                else {
                    alert(strbundle.getString("alert.error.calendar.file.read") + "\n" +
                            file.path);
                }
			}
            else {
                alert(strbundle.getString("alert.error.calendar.file.open") + "\n" +
                        file.path);
            }
        }
    },*/

    // Add an event to the tree.

    addTreeRow: function(treeChildren, data) {
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);
        var item = document.createElement("treeitem");
        var row = document.createElement("treerow");
        var cell;

        // First cell is "startDateDisplay", not editable.
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.startDateDisplay);
        cell.setAttribute("value", data.startDate);
        cell.setAttribute("startTimeMilli", data.startTimeMilli);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Second cell is "timeDisplay", not editable. Stash "time24hr" and "startTimeMilli".
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.timeDisplay);
        cell.setAttribute("value", data.time24hr);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Third cell is "endDateDisplay", not editable.
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.endDateDisplay);
        cell.setAttribute("value", data.endDate);
        cell.setAttribute("endTimeMilli", data.endTimeMilli);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Fourth cell is "recurring", displayed as a checkbox.
        cell = document.createElement("treecell");
        cell.setAttribute("value", data.recurring);
        row.appendChild(cell);

        // Fifth cell is "freq".
        cell = document.createElement("treecell");
        cell.setAttribute("label", this.getFreq(data.repeatRule.freq));
        cell.setAttribute("value", json.encode(data.repeatRule));
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Sixth cell is "repeat until".
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.untilDateDisplay);

        // Save both in value for easy access when toggling Recurring.
        cell.setAttribute("value", json.encode({untilDate: data.untilDate, untilTimeMilli: data.untilTimeMilli}));
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Seventh cell is "alertDays".
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.alertDays);
        cell.setAttribute("value", data.alertMilli);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Eighth cell is "descr".
        cell = document.createElement("treecell");
        cell.setAttribute("label", data.description);
        cell.setAttribute("editable", "false");
        row.appendChild(cell);

        // Add the treerow onto treeitem.
        item.appendChild(row);

        // Append or replace? Check our custom attrribute.
        // When called from onLoad, updaterow = "".
        var updaterow = document.getElementById("simclk-tboxEventDesc").
                getAttribute("updaterow");

        if ( updaterow ) {
            treeChildren.replaceChild(item, treeChildren.childNodes[parseInt(updaterow, 10)]);
            document.getElementById("simclk-tboxEventDesc").setAttribute("updaterow", "");
        }
        else {
            treeChildren.appendChild(item);
        }
    },

    // Called when Delete button is clicked.

    onDelete: function() {
        var strbundle = document.getElementById("simclk-strings");
        var tree = document.getElementById("simclk-treeEvents");

        if ( tree.currentIndex < 0 ) {
            // User didn't make a selection.
            alert(strbundle.getString("alert.warning.no.item.selected"));
        }
        else {
            var start = {};
            var end = {};
            var treeItem;
            var numRanges = tree.view.selection.getRangeCount();

            // Start with the last selection range and work up.
            for ( var i = numRanges - 1; i >= 0; i-- ) {
                tree.view.selection.getRangeAt(i,start,end);

                // Within a range, delete from the bottom up.
                for ( var j = end.value; j >= start.value; j-- ) {
                    treeItem = tree.view.getItemAtIndex(j);
                    treeItem.parentNode.removeChild(treeItem);
                }
            }
        }
    },

    // Called when Del All button is clicked.

    onDeleteAll: function() {
        var treeChildren = document.getElementById("simclk-treeItems");

        // Remove from the bottom.
        while ( treeChildren.hasChildNodes() ) {
            treeChildren.removeChild(treeChildren.lastChild);
        }
    },

    // Called when user clicks OK.

    onOK: function() {
        this.buildCalEventTable();

        // Stealth mode updates the tables, but does not issue an alert.
        this.checkCalEvents("stealth");
        this.updateCalEventFile();
    },

    // Retrieve the global. Could be complete table, today's events or upcoming.

    getCalEventTable: function(eventTableStr) {
        var strbundle = document.getElementById("simclk-strings");
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);
        var calEventTable = [];

        try {
            if ( Application.storage.has(eventTableStr) ) {
                 calEventTable = Application.storage.get(eventTableStr, "ERR");

                if ( calEventTable === "ERR" ) {
                    alert(strbundle.getString("alert.error.loading.events"));
                    return false;
                }
            }
            else {
                Application.storage.set(eventTableStr, calEventTable);
            }
        }
        catch(e) {
            alert(strbundle.getString("alert.error.loading.events") +
                  "\n" + e.message);
            return false;
        }

        return calEventTable;
    },

    // Update the global.

    setCalEventTable: function(eventTable, eventTableStr) {
        var Application = Components.classes["@mozilla.org/fuel/application;1"].
                getService(Components.interfaces.fuelIApplication);

        Application.storage.set(eventTableStr, eventTable);
    },

    // Construct an array of event objects.
    // Called when the user clicks OK or Add/OK.

    buildCalEventTable: function() {
        var json = Components.classes["@mozilla.org/dom/json;1"].
                createInstance(Components.interfaces.nsIJSON);
        var treeChildren = document.getElementById("simclk-treeItems");
        var calEventTable = [];

        if ( treeChildren.hasChildNodes() ) {
            // Build table of all events from dialog Events tree.
            var row;
            var recurring, startDateDisplay, startDate, startTimeMilli, endDateDisplay, endDate,
                endTimeMilli, timeDisplay, time24hr, freq, repeatRule, until, untilDateDisplay,
                untilDate, untilTimeMilli, alertDays, alertMilli, description;
            var startDateNode, endDateNode, timeNode, recurringNode, freqNode,
                untilNode, alertDaysNode, descrNode = null;

            var treeItems = treeChildren.childNodes;
            var len = treeItems.length;

            for ( var i = 0; i < len; i++ ) {
                row = treeItems[i].firstChild;

                startDateNode = row.firstChild;
                startDateDisplay = startDateNode.getAttribute("label");
                startDate = startDateNode.getAttribute("value");
                startTimeMilli = parseInt(startDateNode.getAttribute("startTimeMilli"), 10);

                timeNode = startDateNode.nextSibling;
                timeDisplay = timeNode.getAttribute("label");
                time24hr = timeNode.getAttribute("value");

                endDateNode = timeNode.nextSibling;
                endDateDisplay = endDateNode.getAttribute("label");
                endDate = endDateNode.getAttribute("value");
                endTimeMilli = parseInt(endDateNode.getAttribute("endTimeMilli"), 10);

                // Convert string attribute to boolean.
                recurringNode = endDateNode.nextSibling;
                recurring = ( recurringNode.getAttribute("value") === "true" ) ?
                        true : false;

                freqNode = recurringNode.nextSibling;
                repeatRule = json.decode(freqNode.getAttribute("value"));

                untilDateNode = freqNode.nextSibling;
                until = json.decode(untilDateNode.getAttribute("value"));
                untilDateDisplay = untilDateNode.getAttribute("label");

                alertDaysNode = untilDateNode.nextSibling;
                alertDays = parseInt(alertDaysNode.getAttribute("label"), 10);
                alertMilli = parseInt(alertDaysNode.getAttribute("value"), 10);

                descrNode = alertDaysNode.nextSibling;
                description = descrNode.getAttribute("label");

                calEventTable[i] = {recurring: recurring,
                                    startDateDisplay: startDateDisplay,
                                    startDate: startDate,
                                    startTimeMilli: startTimeMilli,
                                    endDateDisplay: endDateDisplay,
                                    endDate: endDate,
                                    endTimeMilli: endTimeMilli,
                                    timeDisplay: timeDisplay,
                                    time24hr: time24hr,
                                    repeatRule: repeatRule,
                                    untilDateDisplay: untilDateDisplay,
                                    untilDate: until.untilDate,
                                    untilTimeMilli: until.untilTimeMilli,
                                    alertDays: alertDays,
                                    alertMilli: alertMilli,
                                    description: description
                                   };
            }
        }

        // Sort the table on startTimeMilli property.
        if ( calEventTable.length > 1 ) {
            calEventTable.sort(this.sortCalEventTable);
        }

        this.setCalEventTable(calEventTable, "calEventTable");
    },

    // Construct an array of calendar event objects from the calendar event file,
    // and store via FUEL. Called at startup.

    buildCalEventTableFromFile: function() {
        var strbundle = document.getElementById("simclk-strings");

        try {
            var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService).
                    getBranch("extensions.simpletimerClocks@grbradt.org.");
            var file = prefs.getComplexValue("myDir",
                    Components.interfaces.nsILocalFile);
        }
        catch(e) {
            alert(strbundle.getString("alert.error.get.prefs") +
                    "\n" + e.message);
            return;
        }

        file.append("calEvents.txt");
        var calEventTable = [];

        if ( file.exists() ) {
            var inData = SimpleClocksFileIO.read(file);

            if ( inData ) {
                // Objects stored as JSON strings in the pref.
                var json = Components.classes["@mozilla.org/dom/json;1"].
                        createInstance(Components.interfaces.nsIJSON);

                var calEventArray = inData.split(this.calEventSep);

                for ( var i in calEventArray ) {
                    calEventTable[i] = json.decode(calEventArray[i]);
                }

                // Sort the table on startTimeMilli.
                if ( calEventTable.length > 1 ) {
                    calEventTable.sort(this.sortCalEventTable);
                }
            }
        }

        this.setCalEventTable(calEventTable, "calEventTable");
    },

    // Write calendar events file from event table.

    updateCalEventFile: function() {
        var strbundle = document.getElementById("simclk-strings");

        try {
            var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService).
                    getBranch("extensions.simpletimerClocks@grbradt.org.");
            var file = prefs.getComplexValue("myDir",
                    Components.interfaces.nsILocalFile);
        }
        catch(e) {
            alert(strbundle.getString("alert.error.get.prefs") +
                    "\n" + e.message);
            return;
        }

        var bkup = file.clone();
        bkup.append("calEvents-bkup.txt");
        file.append("calEvents.txt");

        if ( file.exists() ) {
            var calEventTable = this.getCalEventTable("calEventTable");

            // Objects stored as JSON strings in the pref.
            var json = Components.classes["@mozilla.org/dom/json;1"].
                    createInstance(Components.interfaces.nsIJSON);

            var outData = "";

            for ( var i in calEventTable ) {
                outData += json.encode(calEventTable[i]) + this.calEventSep;
            }

            // Remove last separator.
            if ( outData ) {
                outData = outData.substring(0, outData.length - 1);
            }

            // Overwrite calendar file.
            if ( SimpleClocksFileIO.write(file, outData, "") &&
                 SimpleClocksFileIO.write(bkup, outData, "")) {
                if ( calEventTable.length > 0 ) {
                    prefs.setCharPref("calEventsUpdated", new Date().toLocaleString());
                }
                else {
                    prefs.setCharPref("calEventsUpdated", "");
                }
            }
            else {
                alert(strbundle.getString("alert.error.calendar.file.write") + "\n" +
                        file.path);
            }
        }
        else {
            alert(strbundle.getString("alert.error.calendar.file.open") + "\n" +
                    file.path);
        }
    },

    // Sort the calEvent table on objects startTimeMilli property.

    sortCalEventTable: function(a, b) {
        var x = a.startTimeMilli;
        var y = b.startTimeMilli;

        return ( (x < y) ? -1 : ( (x > y) ? 1 : 0 ) );
    },

    // Check for upcoming or today's events.

    checkCalEvents: function(mode) {
        this.debug("Simple Timer + Clocks: check calendar events at " + new Date().toLocaleString());
        var strbundle = document.getElementById("simclk-strings");

        try {
            var prefs = Components.classes["@mozilla.org/preferences-service;1"].
                    getService(Components.interfaces.nsIPrefService).
                    getBranch("extensions.simpletimerClocks@grbradt.org.");
        }
        catch(e) {
            alert(strbundle.getString("alert.error.get.prefs") +
                    "\n" + e.message);
            return;
        }

        var currDate = new Date();

        // String comparison; datepicker uses 1-12 for month.
        var currDateStr = currDate.getFullYear() + "-" +
                SimpleClocksCalendar.pad(currDate.getMonth() + 1) + "-" +
                SimpleClocksCalendar.pad(currDate.getDate());
        var currTimeMilli = currDate.getTime();
        var todayEvents = [];
        var upcomingEvents = [];
        var startDate, endDate, newStartDate, newEndDate, untilDate = null;
        var status = "";
        var newStartDateStr, diffMilli, oldDate, dow, nth;
        var logStartDate, logUntilDate, logStartTime, logRecurring, logDescription;

        if ( !SimpleClocks.midnightCalendarID && currDate.getHours() > (23 - SimpleClocks.calendarAlertHours) ) {
            // Check at midnight.
            var midnight = new Date();
            midnight.setDate(midnight.getDate() + 1);
            midnight.setHours(0,0,1,0);
            var delay = midnight.getTime() - currTimeMilli;

            // Stealth mode updates the tables, but does not issue an alert.
            SimpleClocks.midnightCalendarID = setTimeout("SimpleClocksCalendar.checkCalEvents('stealth')",
                                                         delay);
        }

        if ( SimpleClocks.midnightCalendarID && currDate.getHours() < (23 - SimpleClocks.calendarAlertHours) ) {
            // Midnight timer has triggered.
            SimpleClocks.midnightCalendarID = 0;
        }

        var wm = Components.classes["@mozilla.org/appshell/window-mediator;1"].getService();
        var wmed = wm.QueryInterface(Components.interfaces.nsIWindowMediator);
        var enumerator = wmed.getEnumerator("navigator:browser");

        // Only check for one (oldest) window.
        while ( enumerator.hasMoreElements() ) {
            var win = enumerator.getNext();

            // Compare window references.
            if ( win.window != window.window ) {
                return;
            }

            break;
        }

        // For comparison (event dates use 0:00:00).
        currDate.setHours(0,0,0,0);

        var calEventTable = SimpleClocksCalendar.getCalEventTable("calEventTable");

        for ( var i = 0; i < calEventTable.length; i++ ) {
            if ( currDateStr < calEventTable[i].startDate ) {
                if ( calEventTable[i].alertMilli > 0 &&
                    currTimeMilli > calEventTable[i].alertMilli ) {
                    // Upcoming.
                    upcomingEvents.push(calEventTable[i]);
                }
            }
            else if ( currDateStr <= calEventTable[i].endDate ) {
                // Today's event.
                todayEvents.push(calEventTable[i]);
            }
            else {
                // Event completed. Either remove or update.
                // First save info for logging.
                if ( SimpleClocks.eventLogging ) {
                    // The calEventTable[i] object may be deleted.
                    logStartDate = calEventTable[i].startDateDisplay;
                    logUntilDate = calEventTable[i].untilDateDisplay;
                    logStartTime = calEventTable[i].timeDisplay;
                    logRecurring = ( calEventTable[i].recurring ) ?
                            SimpleClocksCalendar.getFreq(calEventTable[i].repeatRule.freq) :
                            strbundle.getString("msg.no");
                    logDescription = calEventTable[i].description;
                }

                if ( calEventTable[i].recurring &&
                   ( calEventTable[i].untilDateDisplay === strbundle.getString("display.forever") ||
                     currTimeMilli < calEventTable[i].untilTimeMilli ) ) {
                    // Update. Set a new event start date.
                    // status used for logging.
                    status = strbundle.getString("msg.completed");

                    // We need to update event end date too.
                    diffMilli = calEventTable[i].endTimeMilli - calEventTable[i].startTimeMilli;

                    newStartDate = new Date(calEventTable[i].startTimeMilli);

                    if ( calEventTable[i].repeatRule.freq === 0 ) {
                        // Yearly.
                        oldDate = newStartDate.getDate();

                        while ( currDate > newStartDate ) {
                            newStartDate.setFullYear(newStartDate.getFullYear() + 1);
                        }

                        if ( newStartDate.getDate() !== oldDate ) {
                            // oldDate must be 29 (leap day), and newStartDate is Mar 1.
                            // Set newStartDate to Feb 28.
                            newStartDate.setDate(0);
                        }
                    }
                    else if ( calEventTable[i].repeatRule.freq === 1 ) {
                        // Yearly by day.
                        dow = newStartDate.getDay();
                        nth = parseInt(calEventTable[i].repeatRule.byDay, 10);
                        oldDate = newStartDate.getDate();

                        while ( currDate > newStartDate ) {
                            newStartDate.setFullYear(newStartDate.getFullYear() + 1);
                        }

                        if ( newStartDate.getDate() !== oldDate ) {
                            // oldDate must be 29 (leap day), and newStartDate is Mar 1.
                            // Set newStartDate to Feb 28.
                            newStartDate.setDate(0);
                        }

                        this.nthDayToDate(newStartDate, nth, dow);
                    }
                    else if ( calEventTable[i].repeatRule.freq === 2 ) {
                        // Monthly.
                        oldDate = newStartDate.getDate();

                        while ( currDate > newStartDate ) {
                            newStartDate.setMonth(newStartDate.getMonth() + 1);
                        }

                        if ( newStartDate.getDate() !== oldDate ) {
                            // eg Jan 31 => Mar 3.
                            // Set newStartDate to last day of previous month.
                            newStartDate.setDate(0);
                        }
                    }
                    else if ( calEventTable[i].repeatRule.freq === 3 ) {
                        // Monthly by day.
                        dow = newStartDate.getDay();
                        nth = parseInt(calEventTable[i].repeatRule.byDay, 10);
                        oldDate = newStartDate.getDate();

                        while ( currDate > newStartDate ) {
                            newStartDate.setMonth(newStartDate.getMonth() + 1);
                        }

                        if ( newStartDate.getDate() !== oldDate ) {
                            // eg Jan 31 => Mar 3.
                            // Set newStartDate to last day of previous month.
                            newStartDate.setDate(0);
                        }

                        this.nthDayToDate(newStartDate, nth, dow);
                    }
                    else if ( calEventTable[i].repeatRule.freq === 4 ) {
                        // Weekly.
                        while ( currDate > newStartDate ) {
                            newStartDate.setDate(newStartDate.getDate() + 7);
                        }
                    }
                    else {
                        // Daily.
                        while ( currDate > newStartDate ) {
                            newStartDate.setDate(newStartDate.getDate() + 1);
                        }
                    }

                    newStartDateStr = newStartDate.getFullYear() + "-" +
                            SimpleClocksCalendar.pad(newStartDate.getMonth() + 1) + "-" +
                            SimpleClocksCalendar.pad(newStartDate.getDate());

                    // Check if the new event start date exceeds a valid until date.
                    if ( calEventTable[i].untilDateDisplay !== strbundle.getString("display.forever") &&
                         newStartDateStr > calEventTable[i].untilDate ) {
                        // Remove.
                        status = strbundle.getString("msg.expired");
                        calEventTable.splice(i, 1);
                        i--;
                    }
                    else {
                        // Update start date.
                        calEventTable[i].startDateDisplay = newStartDate.toLocaleFormat("%b %d %Y");
                        calEventTable[i].startTimeMilli = newStartDate.getTime();
                        calEventTable[i].startDate = newStartDateStr;

                        // Update end date.
                        if ( diffMilli > 0 ) {
                            newEndDate = new Date(calEventTable[i].startTimeMilli + diffMilli);
                            calEventTable[i].endDateDisplay = newEndDate.toLocaleFormat("%b %d %Y");
                            calEventTable[i].endTimeMilli = newEndDate.getTime();
                            calEventTable[i].endDate = newEndDate.getFullYear() + "-" +
                                    SimpleClocksCalendar.pad(newEndDate.getMonth() + 1) + "-" +
                                    SimpleClocksCalendar.pad(newEndDate.getDate());
                        }
                        else {
                            calEventTable[i].endDateDisplay = calEventTable[i].startDateDisplay;
                            calEventTable[i].endTimeMilli = calEventTable[i].startTimeMilli;
                            calEventTable[i].endDate = calEventTable[i].startDate;
                        }

                        if ( calEventTable[i].alertDays > 0 ) {
                            // We can fiddle with newStartDate now.
                            newStartDate.setDate(newStartDate.getDate() - calEventTable[i].alertDays);
                            calEventTable[i].alertMilli = newStartDate.getTime();
                        }

                        // Daily events always alertable.
                        if ( calEventTable[i].repeatRule.freq === 5 ) {
                            todayEvents.push(calEventTable[i]);
                        }
                    }
                }
                else {
                    // Not recurring, or until date exceeded. Remove.
                    status = strbundle.getString("msg.expired");
                    calEventTable.splice(i, 1);
                    i--;
                }

                if ( SimpleClocks.eventLogging ) {
                    // Pass event type, eventStartDate, eventEndDate, event time, recurring, status, description.
					SimpleClocksEventLog.logEvent(strbundle.getString("msg.calendar"),
												  logStartDate,
												  logUntilDate,
                                                  logStartTime,
												  logRecurring,
												  status,
												  logDescription);
				}
            }
        }

        if ( (todayEvents.length > 0 || upcomingEvents.length > 0) &&
              mode === "alert" ) {
            var params = { todayItems: todayEvents, upcomingItems: upcomingEvents, mode: "calEvent" };
			setTimeout(SimpleClocksSliderAlert.addMessageToQueue, 5000, params);
			setTimeout("SimpleClocks.playSound(3)", 5000);
        }

        // Sort the tables on startTimeMilli property.
        if ( todayEvents.length > 1 ) {
            todayEvents.sort(this.sortCalEventTable);
        }

        if ( upcomingEvents.length > 1 ) {
            upcomingEvents.sort(this.sortCalEventTable);
        }

        SimpleClocksCalendar.setCalEventTable(todayEvents, "todayEventsTable");
        SimpleClocksCalendar.setCalEventTable(upcomingEvents, "upcomingEventsTable");

        if ( status ) {
            if ( calEventTable.length > 1 ) {
                calEventTable.sort(this.sortCalEventTable);
            }

            SimpleClocksCalendar.setCalEventTable(calEventTable, "calEventTable");

            // This method will trigger table updates in main.
            SimpleClocksCalendar.updateCalEventFile();
        }
        else {
            // Trigger table updates in main.
            prefs.setCharPref("calEventsUpdated", new Date().toLocaleString());
        }
    },

    // Convert numeric value to recurring frequency.

    getFreq: function(freq) {
        var strbundle = document.getElementById("simclk-strings");
        var frequencies = [strbundle.getString("display.freq.yearly"),
                           strbundle.getString("display.freq.yearly.by.day"),
                           strbundle.getString("display.freq.monthly"),
                           strbundle.getString("display.freq.monthly.by.day"),
                           strbundle.getString("display.freq.weekly"),
                           strbundle.getString("display.dow.daily")
                          ];

        if ( frequencies[freq] != null ) {
            // When undefined (ie. frequencies[-1]) is compared to null, the result is true.
            return frequencies[freq];
        }
        else if ( freq === -1 ) {
            return strbundle.getString("msg.not.applicable");
        }
        else {
            alert(strbundle.getString("alert.error.invalid.freq"));
            return "Err";
        }
    },

    // Pad a number with zero if necessary.

    pad: function(num) {
        return ( ( num < 10 ) ? "0" + num : num );
    },

    // Called on startup. Create the file and save the path.

    createCalendarEventsFile: function() {
        var strbundle = document.getElementById("simclk-strings");
        var file = Components.classes["@mozilla.org/file/directory_service;1"].
                getService(Components.interfaces.nsIProperties).
                get("ProfD", Components.interfaces.nsIFile);

        file.append("SimpleTimerClocks");

        if ( !file.exists() ) {
            SimpleClocks.prefs.setComplexValue("myDir", Components.interfaces.nsILocalFile, file);
        }

        var bkup = file.clone();
        file.append("calEvents.txt");
        bkup.append("calEvents-bkup.txt");

        if ( !file.exists() ) {
            if ( !SimpleClocksFileIO.create(file) ) {
                alert(strbundle.getString("alert.error.calendar.file.create") + "\n" +
                        file.path);
                return false;
            }
        }

        if ( !bkup.exists() ) {
            if ( !SimpleClocksFileIO.create(bkup) ) {
                alert(strbundle.getString("alert.error.calendar.file.create") + "\n" +
                        bkup.path);
                return false;
            }
        }

        return true;
    },

    // Dump to console.

    dumpCalEventObject: function(eventObj) {
        this.debug("\nEvent data:");
        this.debug("Event start date: " + eventObj.startDateDisplay);
        this.debug("Event start date (datepicker format): " + eventObj.startDate);
        this.debug("Event start date (millisecs): " + eventObj.startTimeMilli);
        this.debug("Event display time: " + eventObj.timeDisplay);
        this.debug("Event time (timepicker format): " + eventObj.time24hr);
        this.debug("Event end date: " + eventObj.endDateDisplay);
        this.debug("Event end date (datepicker format): " + eventObj.endDate);
        this.debug("Event end date (millisecs): " + eventObj.endTimeMilli);
        this.debug("Recurring?: " + eventObj.recurring);
        this.debug("Event frequency: " + eventObj.repeatRule.freq);
        this.debug("Event byDay: " + eventObj.repeatRule.byDay);
        this.debug("Event until date: " + eventObj.untilDateDisplay);
        this.debug("Event until date (datepicker format): " + eventObj.untilDate);
        this.debug("Event until date (millisecs): " + eventObj.untilTimeMilli);
        this.debug("Alert days prior to event: " + eventObj.alertDays);
        this.debug("Alert time (millisecs): " + eventObj.alertMilli);
        this.debug("Description: " + eventObj.description);
    },

    // Debug messages to console.

    debug: function(aMsg) {
        setTimeout(function() { throw new Error("[debug] " + aMsg); }, 0);
    }
};
